with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics.long_elementary_functions,
  terms,
  nlp;

procedure istress is

  subtype fmat3x1 is terms.fmat1 (1 .. 3);

  subtype fmat3x3 is terms.fmat2 (1 .. 3, 1 .. 3);

  type neo_rec is
    record
      mu : long_float;
    end record;

  type mooney_rec is
    record
      mu10 : long_float;
      mu01 : long_float;
    end record;

  type boyce_rec is
    record
      mu     : long_float;
      lambda : long_float;
      c      : terms.fmat1 (1 .. 5);
    end record;

  type elem2_rec is
    record
      area : long_float;
      v1v1 : long_float;
      v1v2 : long_float;
      v2v2 : long_float;
      glob : terms.imat1 (1 .. 3);
      v1   : terms.fmat1 (1 .. 3);
      v2   : terms.fmat1 (1 .. 3);
      w    : terms.fmat1 (1 .. 3);
    end record;

  type elem2_vec is array (integer range <>) of elem2_rec;

  type elem2_ptr is access elem2_vec;

  optimal                           : boolean;
  nn,n,maxls,minor,maxminor,maxlist : integer;
  lstol,mntol,mnref,maxnorm,f       : long_float;
  thickness                         : long_float;
  index                             : terms.imat2_ptr;
  xo,yo,zo,x,g                      : terms.fmat1_ptr;
  known,force                       : terms.fmat2_ptr;
  mat                               : integer;
  neo                               : neo_rec;
  mooney                            : mooney_rec;
  boyce                             : boyce_rec;
  elem2                             : elem2_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    e,ne,glob,axis,nk : integer;
    io                : file_type;

  begin
    open(io, in_file, "input.txt");

    -- lstol: Controls the accuracy of the line search. It must
    -- lie in the range 0.0 <= lstol < 1.0. Decreasing this value
    -- tends to increase the accuracy of the line search. It is
    -- usually 0.1.

    -- maxls: The maximum number of cubic interpolations allowed
    -- in the line search. It must be greater that one. It is
    -- usually 20.

    -- The iterations terminate if the infinity norm of the
    -- residue becomes less than or equal to maxnorm, which is
    -- calculated as:

    -- maxnorm = mntol * mnref

    -- If mntol is specified as a non-positive value, it is
    -- replaced by 0.001. If mnref is specified as a non-positive
    -- value, it is replaced by the infinite norm of the residue,
    -- evaluated with all unknown displacements equal to zero.

    -- maxminor: The maximum number of iterations allowed. It must
    -- be greater that zero. It is usually ten times the number of
    -- degrees of freedom.

    -- maxlist: The number of BFGS corrections kept. These
    -- corrections are stored separately on a doubly linked
    -- circular list. Used only with the limited memory BFGS
    -- method. It is usually between 3 and 20.

    scan(io);
    get(io,lstol);
    get(io,maxls);
    get(io,mntol);
    get(io,mnref);
    get(io,maxminor);
    get(io,maxlist);
    skip_line(io);

    if mntol <= 0.0 then
      mntol := 0.001;
    end if;

    -- thickness
    scan(io);
    get(io,thickness);

    -- material
    get(io,mat);

    case mat is
      when 1 =>
        get(io,neo.mu);

      when 2 =>
        get(io,mooney.mu10);
        get(io,mooney.mu01);

      when 3 =>
        get(io,boyce.mu);
        get(io,boyce.lambda);
        boyce.c(1) := 1.0 / 2.0;
        boyce.c(2) := 1.0 / 20.0 / boyce.lambda ** 2;
        boyce.c(3) := 11.0 / 1050.0 / boyce.lambda ** 4;
        boyce.c(4) := 19.0 / 7000.0 / boyce.lambda ** 6;
        boyce.c(5) := 519.0 / 673750.0 / boyce.lambda ** 8;

      when others =>
        null;
    end case;

    skip_line(io);

    -- nodes
    scan(io);
    get(io,nn);
    skip_line(io);

    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    skip_line(io);
    for k in 1 .. nn loop
      get(io,glob);
      get(io,xo(glob));
      get(io,yo(glob));
      get(io,zo(glob));
      skip_line(io);
    end loop;

    -- elements 2
    scan(io);
    get(io,ne);
    skip_line(io);

    elem2 := new elem2_vec (1 .. ne);

    skip_line(io);
    for k in elem2'range loop
      get(io,e);
      get(io,elem2(e).glob(1));
      get(io,elem2(e).glob(2));
      get(io,elem2(e).glob(3));
      skip_line(io);
    end loop;

    -- initiate index
    index := new terms.imat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        index(glob,axis) := 0;
      end loop;
    end loop;

    -- initiate known
    known := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        known(glob,axis) := 0.0;
      end loop;
    end loop;

    -- known displacements
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,known(glob,axis));
        skip_line(io);
        index(glob,axis) := 1;
      end loop;
    end if;

    -- initiate force
    force := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := 0.0;
      end loop;
    end loop;

    -- nodal forces
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,force(glob,axis));
        skip_line(io);
      end loop;
    end if;

    close(io);
  end input;

  procedure element2 (elem2                 : in elem2_rec;
                      fe,r1,c1,c2,psi1,psi2 : out long_float;
                      x                     : in terms.fmat1_ptr;
                      ge                    : out fmat3x3) is

    use
      ada.numerics.long_elementary_functions;

    glob,i                        : integer;
    area,v1z2,v2z1,psi            : long_float;
    v1v1,v1v2,v2v2,z1z1,z1z2,z2z2 : long_float;
    a1,a2,a3,a4,r2,r3,r4,r5       : long_float;
    u1,u2,u3,v1,v2,w,w1,w2,z1,z2  : terms.fmat1 (1 .. 3);

  begin
    -- u1
    glob := elem2.glob(1);
    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        u1(axis) := x(i);
      else
        u1(axis) := known(glob,axis);
      end if;
    end loop;

    -- u2
    glob := elem2.glob(2);
    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        u2(axis) := x(i);
      else
        u2(axis) := known(glob,axis);
      end if;
    end loop;

    -- u3
    glob := elem2.glob(3);
    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        u3(axis) := x(i);
      else
        u3(axis) := known(glob,axis);
      end if;
    end loop;

    -- v1
    for axis in 1 .. 3 loop
      v1(axis) := elem2.v1(axis) + u3(axis) - u2(axis);
    end loop;

    -- v2
    for axis in 1 .. 3 loop
      v2(axis) := elem2.v2(axis) + u1(axis) - u3(axis);
    end loop;

    -- w = v1 * v2
    w(1) := v1(2) * v2(3) - v1(3) * v2(2);
    w(2) := v1(3) * v2(1) - v1(1) * v2(3);
    w(3) := v1(1) * v2(2) - v1(2) * v2(1);

    -- area
    area := 0.0;
    for axis in 1 .. 3 loop
      area := area + w(axis) ** 2;
    end loop;
    area := sqrt(area);

    -- w
    for axis in 1 .. 3 loop
      w(axis) := w(axis) / area;
    end loop;

    -- w1 = w x v1
    w1(1) := w(2) * v1(3) - w(3) * v1(2);
    w1(2) := w(3) * v1(1) - w(1) * v1(3);
    w1(3) := w(1) * v1(2) - w(2) * v1(1);

    -- w2 = w x v2
    w2(1) := w(2) * v2(3) - w(3) * v2(2);
    w2(2) := w(3) * v2(1) - w(1) * v2(3);
    w2(3) := w(1) * v2(2) - w(2) * v2(1);

    -- z1
    for axis in 1 .. 3 loop
      z1(axis) := elem2.v1v1 * v2(axis) - elem2.v1v2 * v1(axis);
    end loop;

    -- z2
    for axis in 1 .. 3 loop
      z2(axis) := elem2.v1v2 * v2(axis) - elem2.v2v2 * v1(axis);
    end loop;

    -- a1, a2, a3, a4
    a1 := elem2.area;
    a2 := a1 * a1;
    a3 := a2 * a1;
    a4 := a3 * a1;

    -- r1, r2, r3, r4, r5
    r1 := elem2.area / area;
    r2 := r1 * r1;
    r3 := r2 * r1;
    r4 := r3 * r1;
    r5 := r4 * r1;

    -- invariant 1
    v1z2 := 0.0;
    v2z1 := 0.0;
    for axis in 1 .. 3 loop
      v1z2 := v1z2 + v1(axis) * z2(axis);
      v2z1 := v2z1 + v2(axis) * z1(axis);
    end loop;

    c1 := (v2z1 - v1z2) / a2 + r2;

    -- invariant 2
    v1v1 := 0.0;
    v1v2 := 0.0;
    v2v2 := 0.0;
    z1z1 := 0.0;
    z1z2 := 0.0;
    z2z2 := 0.0;
    for axis in 1 .. 3 loop
      v1v1 := v1v1 + v1(axis) * v1(axis);
      v1v2 := v1v2 + v1(axis) * v2(axis);
      v2v2 := v2v2 + v2(axis) * v2(axis);
      z1z1 := z1z1 + z1(axis) * z1(axis);
      z1z2 := z1z2 + z1(axis) * z2(axis);
      z2z2 := z2z2 + z2(axis) * z2(axis);
    end loop;

    c2 := (v1v1 * z2z2 - 2.0 * v1v2 * z1z2 + v2v2 * z1z1) / a4 + r4;

    -- strain energy per unit undeformed volume
    case mat is
      when 1 =>
        psi := neo.mu * c1 / 2.0;
        psi1 := neo.mu / 2.0;
        psi2 := 0.0;

      when 2 =>
        psi := mooney.mu10 * c1 + mooney.mu01 * (c1 * c1 - c2) / 2.0;
        psi1 := mooney.mu10 + mooney.mu01 * c1;
        psi2 := -mooney.mu01 / 2.0;

      when 3 =>
        psi := 0.0;
        psi1 := 0.0;
        for k in reverse 1 .. 5 loop
          psi := (boyce.c(k) + psi) * c1;
          psi1 := (long_float(k) * boyce.c(k) + psi1) * c1;
        end loop;
        psi := psi * boyce.mu;
        psi1 := psi1 * boyce.mu;
        psi2 := 0.0;

      when others =>
        null;
    end case;

    -- objective
    fe := psi * a1 * thickness / 2.0;

    -- gradient
    for axis in 1 .. 3 loop
      ge(1,axis) := 0.0;
      ge(2,axis) := 0.0;
      ge(3,axis) := 0.0;
    end loop;

    for axis in 1 .. 3 loop
      ge(1,axis) := ge(1,axis) + psi1 * (z1(axis) / a1 - r3 * w1(axis));
      ge(2,axis) := ge(2,axis) + psi1 * (z2(axis) / a1 - r3 * w2(axis));
    end loop;

    for axis in 1 .. 3 loop
      ge(1,axis) := ge(1,axis) + 2.0 * psi2 * ((z1z1 * v2(axis) - z1z2 * v1(axis)) / a3 - r5 * w1(axis));
      ge(2,axis) := ge(2,axis) + 2.0 * psi2 * ((z1z2 * v2(axis) - z2z2 * v1(axis)) / a3 - r5 * w2(axis));
    end loop;

    for axis in 1 .. 3 loop
      ge(1,axis) := ge(1,axis) * thickness;
      ge(2,axis) := ge(2,axis) * thickness;
      ge(3,axis) := -(ge(1,axis) + ge(2,axis));
    end loop;
  end element2;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

    glob,i                : integer;
    fe,r1,c1,c2,psi1,psi2 : long_float;
    ge                    : fmat3x3;

  begin
    -- initiate objective, gradient
    f := 0.0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i <= n then
          f := f - force(glob,axis) * x(i);
          g(i) := -force(glob,axis);
        end if;
      end loop;
    end loop;

    -- elements 2
    for e in elem2'range loop
      element2(elem2(e),fe,r1,c1,c2,psi1,psi2,x,ge);

      -- update objective
      f := f + fe;

      -- update gradient
      for loc in 1 .. 3 loop
        glob := elem2(e).glob(loc);
        for axis in 1 .. 3 loop
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + ge(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;
  end objective;

  package qn is new nlp.qnewton(objective);

  procedure start is

    use
      ada.numerics.long_elementary_functions;

    glob1,glob2,glob3,s1,s2 : integer;

  begin
    -- elements 2
    for e in elem2'range loop
      glob1 := elem2(e).glob(1);
      glob2 := elem2(e).glob(2);
      glob3 := elem2(e).glob(3);

      -- v1
      elem2(e).v1(1) := xo(glob3) - xo(glob2);
      elem2(e).v1(2) := yo(glob3) - yo(glob2);
      elem2(e).v1(3) := zo(glob3) - zo(glob2);

      -- v2
      elem2(e).v2(1) := xo(glob1) - xo(glob3);
      elem2(e).v2(2) := yo(glob1) - yo(glob3);
      elem2(e).v2(3) := zo(glob1) - zo(glob3);

      -- v1v1
      elem2(e).v1v1 := 0.0;
      for axis in 1 .. 3 loop
        elem2(e).v1v1 := elem2(e).v1v1 + elem2(e).v1(axis) * elem2(e).v1(axis);
      end loop;

      -- v1v2
      elem2(e).v1v2 := 0.0;
      for axis in 1 .. 3 loop
        elem2(e).v1v2 := elem2(e).v1v2 + elem2(e).v1(axis) * elem2(e).v2(axis);
      end loop;

      -- v2v2
      elem2(e).v2v2 := 0.0;
      for axis in 1 .. 3 loop
        elem2(e).v2v2 := elem2(e).v2v2 + elem2(e).v2(axis) * elem2(e).v2(axis);
      end loop;

      -- w = v1 * v2
      elem2(e).w(1) := elem2(e).v1(2) * elem2(e).v2(3) - elem2(e).v1(3) * elem2(e).v2(2);
      elem2(e).w(2) := elem2(e).v1(3) * elem2(e).v2(1) - elem2(e).v1(1) * elem2(e).v2(3);
      elem2(e).w(3) := elem2(e).v1(1) * elem2(e).v2(2) - elem2(e).v1(2) * elem2(e).v2(1);

      -- area
      elem2(e).area := 0.0;
      for axis in 1 .. 3 loop
        elem2(e).area := elem2(e).area + elem2(e).w(axis) ** 2;
      end loop;
      elem2(e).area := sqrt(elem2(e).area);

      -- w
      for axis in 1 .. 3 loop
        elem2(e).w(axis) := elem2(e).w(axis) / elem2(e).area;
      end loop;
    end loop;

    -- number of unknowns
    n := 0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          n := n + 1;
        end if;
      end loop;
    end loop;

    -- new sequence
    s1 := 0;
    s2 := n;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          s1 := s1 + 1;
          index(glob,axis) := s1;
        else
          s2 := s2 + 1;
          index(glob,axis) := s2;
        end if;
      end loop;
    end loop;

    -- initiate x
    x := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      x(k) := 0.0;
    end loop;

    -- initiate g
    g := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      g(k) := 0.0;
    end loop;

    -- maxnorm
    if mnref <= 0.0 then
      objective(n,f,x,g);
      maxnorm := mntol * nlp.norm(n,g);
    else
      maxnorm := mntol * mnref;
    end if;
  end start;

  procedure putchar (io : in text_io.file_type;
                     c  : in character;
                     n  : in integer) is

  begin
    for k in 1 .. n loop
      text_io.put(io,c);
    end loop;
  end putchar;

  procedure output is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    infinity : constant long_float := 1.0E+20;

    i,glob,last                   : integer;
    fe,r1,c1,c2,psi1,psi2         : long_float;
    gamma,s,p,r,s1,s2,maxs1,mins2 : long_float;
    ge                            : fmat3x3;
    io                            : file_type;

  begin
    create(io, out_file, "output.txt");

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Minor Iterations                            | ");
    put(io,minor,15);
    put(io," |");
    new_line(io);

    put(io,"| Total Potential Energy                      | ");
    put(io,f,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Allowable Maximum Residue                   | ");
    put(io,maxnorm,2,8,3);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- displacements
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Displ X             Displ Y             Displ Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        i := index(glob,axis);
        if i <= n then
          put(io,x(i),2,8,3);
        else
          put(io,known(glob,axis),2,8,3);
        end if;
      end loop;
      new_line(io);
    end loop;

    -- principal stress
    new_line(io);
    new_line(io);
    putchar(io,'-',41);
    new_line(io);
    put(io," Elem            Sigma1            Sigma2");
    new_line(io);
    putchar(io,'-',41);
    new_line(io);

    maxs1 := -infinity;
    mins2 :=  infinity;

    -- support reactions
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := -force(glob,axis);
      end loop;
    end loop;

    -- elements 2
    for e in elem2'range loop
      element2(elem2(e),fe,r1,c1,c2,psi1,psi2,x,ge);

      -- principal stress
      gamma := 2.0 * r1 * r1 * (psi1 + 2.0 * psi2 * r1 * r1);

      s := c2 + c2 - c1 * c1;
      r := (psi1 * c1 + 2.0 * psi2 * c2 - 0.75 * gamma) * gamma;
      r := r + psi1 * psi1 * s;
      r := r + 4.0 * psi1 * psi2 * (s * c1 + 6.0);
      r := r + 4.0 * psi2 * psi2 * (s * c1 + 8.0) * c1;
      r := sqrt(r);

      s := c1 * c1 - c2;
      p := (3.0 * gamma - 4.0 * (psi1 * c1 + 2.0 * psi2 * c2)) * gamma;
      p := p + 2.0 * psi1 * psi1 * s;
      p := p + 4.0 * psi1 * psi2 * (s * c1 - 6.0);
      p := p + 4.0 * psi2 * psi2 * (s * s - 8.0 * c1);

      s := psi1 * c1 + 2.0 * psi2 * c2 - 1.5 * gamma;

      if s >= 0.0 then
        s1 := s + r;
        s2 := p / s1;
      else
        s2 := s - r;
        s1 := p / s2;
      end if;

      if s1 > maxs1 then
        maxs1 := s1;
      end if;

      if s2 < mins2 then
        mins2 := s2;
      end if;

      put(io,e,5);
      putchar(io,' ',5);
      put(io,s1,2,6,3);
      putchar(io,' ',5);
      put(io,s2,2,6,3);
      new_line(io);

      -- support reactions
      for loc in 1 .. 3 loop
        glob := elem2(e).glob(loc);
        for axis in 1 .. 3 loop
          force(glob,axis) := force(glob,axis) + ge(loc,axis);
        end loop;
      end loop;
    end loop;

    -- principal stress
    putchar(io,'-',41);
    new_line(io);
    put(io,"Max/Min   ");
    put(io,maxs1,2,6,3);
    putchar(io,' ',5);
    put(io,mins2,2,6,3);
    new_line(io);

    -- support reactions
    new_line(io);
    new_line(io);
    putchar(io,'-',59);
    new_line(io);
    put(io," Node           Force X           Force Y           Force Z");
    new_line(io);
    putchar(io,'-',59);
    new_line(io);

    for glob in 1 .. nn loop
      last := 0;
      for axis in 1 .. 3 loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last := axis;
        end if;
      end loop;

      if last /= 0 then
        put(io,glob,5);
        for axis in 1 .. last loop
          if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
            putchar(io,' ',5);
            put(io,force(glob,axis),2,6,3);
          else
            putchar(io,' ',18);
          end if;
        end loop;
        new_line(io);
      end if;
    end loop;

    close(io);
  end output;

  procedure vertex (glob : in integer;
                    vx   : out fmat3x1) is

    i : integer;

  begin
    vx(1) := xo(glob);
    vx(2) := yo(glob);
    vx(3) := zo(glob);

    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        vx(axis) := vx(axis) + x(i);
      else
        vx(axis) := vx(axis) + known(glob,axis);
      end if;
    end loop;
  end vertex;

  procedure autocad is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    vx : fmat3x1;
    io : file_type;

  begin
    create(io, out_file, "autocad.scr");

    -- set running object snap mode to none
    put(io,"osmode 0");
    new_line(io);

    -- elements 2
    put(io,"layer make Initial_2 ");
    new_line(io);

    put(io,"pface");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,xo(glob),1,6,3);
      put(io,",");
      put(io,yo(glob),1,6,3);
      put(io,",");
      put(io,zo(glob),1,6,3);
      new_line(io);
    end loop;
    new_line(io);

    for e in elem2'range loop
      put(io,elem2(e).glob(1),0);
      put(io," ");
      put(io,elem2(e).glob(2),0);
      put(io," ");
      put(io,elem2(e).glob(3),0);
      new_line(io);
      new_line(io);
    end loop;
    new_line(io);

    put(io,"layer make Final_2 ");
    new_line(io);

    put(io,"pface");
    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,vx);
      put(io,vx(1),1,6,3);
      put(io,",");
      put(io,vx(2),1,6,3);
      put(io,",");
      put(io,vx(3),1,6,3);
      new_line(io);
    end loop;
    new_line(io);

    for e in elem2'range loop
      put(io,elem2(e).glob(1),0);
      put(io," ");
      put(io,elem2(e).glob(2),0);
      put(io," ");
      put(io,elem2(e).glob(3),0);
      new_line(io);
      new_line(io);
    end loop;
    new_line(io);

    put(io,"layer off Initial_2 ");
    new_line(io);

    put(io,"layer set 0 ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad;

  procedure ansys is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    nsubst : integer := 10;
    neqit  : integer := 25;
    io     : file_type;

  begin
    create(io, out_file, "ansys.txt");

    new_line(io);
    put(io,"! ENTER THE MODEL CREATION PREPROCESSOR");
    new_line(io);
    put(io,"/PREP7");
    new_line(io);

    new_line(io);
    put(io,"! DEFINE NODES");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,"N, ");
      put(io,glob,5);
      put(io,", ");
      put(io,xo(glob),2,6,3);
      put(io,", ");
      put(io,yo(glob),2,6,3);
      new_line(io);
    end loop;

    new_line(io);
    put(io,"! ACTIVATE DATA TABLE");
    new_line(io);

    case mat is
      when 1 =>
        put(io,"TB, HYPER, 1, , , NEO");
        new_line(io);

        -- mu
        put(io,"TBDATA, 1, ");
        put(io,neo.mu,2,6,3);
        new_line(io);

        -- incompressibility parameter
        put(io,"TBDATA, 2,  0.000000E+00");
        new_line(io);

      when 2 =>
        put(io,"TB, HYPER, 1, , , MOONEY");
        new_line(io);

        -- mu10
        put(io,"TBDATA, 1, ");
        put(io,mooney.mu10,2,6,3);
        new_line(io);

        -- mu01
        put(io,"TBDATA, 2, ");
        put(io,mooney.mu01,2,6,3);
        new_line(io);

         -- incompressibility parameter
        put(io,"TBDATA, 3,  0.000000E+00");
        new_line(io);

      when 3 =>
        put(io,"TB, HYPER, , , , BOYCE");
        new_line(io);

        -- mu
        put(io,"TBDATA, 1, ");
        put(io,boyce.mu,2,6,3);
        new_line(io);

        -- lambda
        put(io,"TBDATA, 2, ");
        put(io,boyce.lambda,2,6,3);
        new_line(io);

         -- incompressibility parameter
        put(io,"TBDATA, 3,  0.000000E+00");
        new_line(io);

      when others =>
        null;
    end case;

    new_line(io);
    put(io,"! DEFINE ELEMENT TYPE");
    new_line(io);
    put(io,"! FULL INTEGRATION WITH B-BAR METHOD");
    new_line(io);
    put(io,"! PLANE STRESS WITH THICKNESS INPUT");
    new_line(io);
    put(io,"ET, 1, PLANE182, 0, , 3");
    new_line(io);

    new_line(io);
    put(io,"! DEFINE REAL CONSTANTS SET");
    new_line(io);
    put(io,"R,     1, ");
    put(io,thickness,2,6,3);
    new_line(io);

    new_line(io);
    put(io,"! DEFINE ELEMENTS");
    new_line(io);

    for e in elem2'range loop
      put(io,"EN, ");
      put(io,e,5);
      put(io,", ");
      put(io,elem2(e).glob(1),5);
      put(io,", ");
      put(io,elem2(e).glob(2),5);
      put(io,", ");
      put(io,elem2(e).glob(3),5);
      put(io,", ");
      put(io,elem2(e).glob(3),5);
      new_line(io);
    end loop;

    new_line(io);
    put(io,"FINISH");
    new_line(io);

    new_line(io);
    put(io,"! ENTER THE SOLUTION PROCESSOR");
    new_line(io);
    put(io,"/SOLU");
    new_line(io);

    new_line(io);
    put(io,"! STATIC ANALYSIS");
    new_line(io);
    put(io,"ANTYPE, STATIC, NEW");
    new_line(io);

    new_line(io);
    put(io,"! DEFINE CONSTRAINTS");
    new_line(io);

    for glob in 1 .. nn loop
      if index(glob,1) > n then
        put(io,"D, ");
        put(io,glob,5);
        put(io,", UX, ");
        put(io,known(glob,1),2,6,3);
        new_line(io);
      end if;

      if index(glob,2) > n then
        put(io,"D, ");
        put(io,glob,5);
        put(io,", UY, ");
        put(io,known(glob,2),2,6,3);
        new_line(io);
      end if;
    end loop;

    new_line(io);
    put(io,"! DELETE FORCE LOADS ON NODES");
    new_line(io);
    put(io,"FDELE, ALL, ALL");
    new_line(io);

    new_line(io);
    put(io,"! SPECIFY FORCES");
    new_line(io);

    for glob in 1 .. nn loop
      if force(glob,1) /= 0.0 then
        put(io,"F, ");
        put(io,glob,5);
        put(io,", FX, ");
        put(io,force(glob,1),2,6,3);
        new_line(io);
      end if;

      if force(glob,2) /= 0.0 then
        put(io,"F, ");
        put(io,glob,5);
        put(io,", FY, ");
        put(io,force(glob,2),2,6,3);
        new_line(io);
      end if;
    end loop;

    new_line(io);
    put(io,"! INCLUDE LARGE DEFORMATION EFFECTS");
    new_line(io);
    put(io,"NLGEOM, ON");
    new_line(io);

    new_line(io);
    put(io,"! SPECIFY NUMBER OF SUBSTEPS");
    new_line(io);
    put(io,"NSUBST, ");
    put(io,nsubst,5);
    new_line(io);

    new_line(io);
    put(io,"! SPECIFY MAXIMUM NUMBER OF EQUILIBRIUM ITERATIONS");
    new_line(io);
    put(io,"NEQIT, ");
    put(io,neqit,5);
    new_line(io);

    new_line(io);
    put(io,"! SET CONVERGENCE VALUES");
    new_line(io);
    put(io,"CNVTOL, F, 1.0, ");
    put(io,maxnorm,2,6,3);
    put(io,", 0");
    new_line(io);

    new_line(io);
    put(io,"! SPECIFY NEWTON-RAPHSON OPTIONS");
    new_line(io);
    put(io,"NROPT, FULL, , OFF");
    new_line(io);

    new_line(io);
    put(io,"! SPECIFY TYPE OF EQUATION SOLVER");
    new_line(io);
    put(io,"EQSLV, FRONT");
    new_line(io);

    new_line(io);
    put(io,"FINISH");
    new_line(io);

    close(io);
  end ansys;

begin
  input;

  start;

  ansys;

  qn.lbfgs(maxlist,n,maxminor,maxls,
           minor,
           maxnorm,lstol,
           f,
           x,g,
           optimal);

  text_io.new_line;
  text_io.put("Minor = ");
  ada.integer_text_io.put(minor,5);
  if optimal then
    text_io.put("     Optimal");
  else
    text_io.put("     NOT Optimal");
  end if;
  text_io.new_line;

  autocad;

  output;
end istress;
